# stdlib imports
import pathlib
from urllib.parse import urljoin

# third party imports
import numpy as np
import pandas as pd
import rasterio
import requests
from bs4 import BeautifulSoup
from rasterio.windows import Window

MAX_INTERFACE_DEPTH = 70  # depth beyond which any tectonic regime has to be intraslab

# Slab 1.0 does not have depth uncertainty, so we make this a constant
DEFAULT_DEPTH_ERROR = 10

ITEM_TEMPLATE = "https://www.sciencebase.gov/catalog/item/{itemid}?format=json"
# ITEM_TEMPLATE = "http://www.sciencebase.gov/catalog/file/get/[ITEMID]"
SEARCH_URL = "https://apps.usgs.gov/shakemap_geodata/slabs/"
KEYWORDS = ["subduction", "model", "slab2"]
SLABGRID_REGEX = "dep|str|dip|unc"


def get_slab_grids(path):
    if not path.exists():
        path.mkdir(parents=True)
    response = requests.get(SEARCH_URL)
    soup = BeautifulSoup(response.text, features="lxml")
    nfiles = 0
    for anchor in soup.find_all("a"):
        if "href" not in anchor.attrs:
            continue
        if ".grd" not in anchor.attrs["href"]:
            continue
        file_name = anchor.attrs["href"]
        file_url = urljoin(SEARCH_URL, file_name)
        response = requests.get(file_url)
        outfile = pathlib.Path(path) / file_name
        with open(outfile, "wb") as fobj:
            fobj.write(response.content)
        nfiles += 1

    return (True, f"Downloaded {nfiles} grid files from {SEARCH_URL} to {path}")


class GridSlab(object):
    """Represents USGS Slab model grids for a given subduction zone."""

    def __init__(self, depth_file, dip_file, strike_file, error_file):
        """Construct GridSlab object from input files.

        Args:
            depth_file (str): Path to Slab depth grid file.
            dip_file (str): Path to Slab dip grid file.
            strike_file (str): Path to Slab strike grid file.
            error_file (str): Path to Slab depth error grid file (can be None).
        """
        self._depth_file = pathlib.Path(depth_file)
        self._dip_file = pathlib.Path(dip_file)
        self._strike_file = pathlib.Path(strike_file)
        self._error_file = pathlib.Path(error_file)

        # there may be a table of maximum slab depths in the same directory
        # as all of the slab grids.  Read it into a local dictionary if found,
        # otherwise we'll use the MAX_INTERFACE_DEPTH constant found above.
        table_file_name = (
            pathlib.Path(__file__).parent / "data" / "maximum_interface_depths.csv"
        )
        if table_file_name.is_file():
            self._slab_table = pd.read_csv(table_file_name)
        else:
            self._slab_table = None

    def getGridValue(self, grid, lat, lon):
        # rasterio does geo-referencing on grid-line basis, we do it on pixel basis
        # (half pixel down and to the right)
        xmax, _ = grid.xy(grid.height, grid.width)
        if xmax > 180:
            lon += 360
        row, col = grid.index(lon, lat)
        nrows, ncols = grid.shape
        if row < 0 or row > nrows - 1 or col < 0 or col > ncols - 1:
            return np.nan

        value = grid.read(1, window=Window(col, row, 1, 1))[0][0]
        return value

    def contains(self, lat, lon):
        """Check to see if input coordinates are contained inside Slab model.

        Args:
            lat (float):  Hypocentral latitude in decimal degrees.
            lon (float):  Hypocentral longitude in decimal degrees.
        Returns:
            bool: True if point falls inside minimum bounding box of slab model.
        """
        with rasterio.open(self._depth_file, "r") as dataset:
            xmin, ymax = dataset.xy(0, 0)
            xmax, ymin = dataset.xy(dataset.height, dataset.width)
            if xmax > 180:
                lon += 360
        if lat >= ymin and lat <= ymax and lon >= xmin and lon <= xmax:
            return True
        return False

    def getSlabInfo(self, lat, lon):
        """Return a dictionary with depth,dip,strike, and depth uncertainty.

        Args:
            lat (float):  Hypocentral latitude in decimal degrees.
            lon (float):  Hypocentral longitude in decimal degrees.
        Returns:
            dict: Dictionary containing keys:
                - region Three letter Slab model region code.
                - strike Slab model strike angle.
                - dip Slab model dip angle.
                - depth Slab model depth (km).
                - depth_uncertainty Slab model depth uncertainty.
        """
        slabinfo = {}
        if not self.contains(lat, lon):
            return slabinfo
        fname = self._depth_file.with_suffix("").name
        parts = fname.split("_")
        region = parts[0]

        with rasterio.open(self._depth_file, "r") as depth_grid:
            # slab grids are negative depth
            depth = -1 * self.getGridValue(depth_grid, lat, lon)

        with rasterio.open(self._dip_file, "r") as dip_grid:
            dip = self.getGridValue(dip_grid, lat, lon)

        with rasterio.open(self._strike_file, "r") as strike_grid:
            strike = self.getGridValue(strike_grid, lat, lon)
            if strike < 0:
                strike += 360

        error = DEFAULT_DEPTH_ERROR
        with rasterio.open(self._error_file, "r") as error_grid:
            error = self.getGridValue(error_grid, lat, lon)
            if np.isnan(strike):
                error = np.nan

        # get the maximum interface depth from table (if present)
        if self._slab_table is not None:
            df = self._slab_table
            max_int_depth = df[df["zone"] == region].iloc[0]["interface_max_depth"]
        else:
            max_int_depth = MAX_INTERFACE_DEPTH

        slabinfo = {
            "region": region,
            "strike": strike,
            "dip": dip,
            "depth": depth,
            "maximum_interface_depth": max_int_depth,
            "depth_uncertainty": error,
        }
        return slabinfo


class SlabCollection(object):
    def __init__(self, datafolder):
        """Object representing a collection of SlabX.Y grids.

        This object can be queried with a latitude/longitude to see if that point is
        within a subduction slab - if so, the slab information is returned.

        Args:
            datafolder (str): String path where grid files and GeoJSON file reside.
        """
        datafolder = pathlib.Path(datafolder)
        self._depth_files = datafolder.glob("*_dep*.grd")

    def getSlabInfo(self, lat, lon, depth):
        """Query the entire set of slab models and return a SlabInfo object, or None.

        Args:
            lat (float):  Hypocentral latitude in decimal degrees.
            lon (float):  Hypocentral longitude in decimal degrees.
            depth (float): Hypocentral depth in km.

        Returns:
            dict: Dictionary containing keys:
                - region Three letter Slab model region code.
                - strike Slab model strike angle.
                - dip Slab model dip angle.
                - depth Slab model depth (km).
                - depth_uncertainty Slab model depth uncertainty.
        """

        deep_depth = 99999999999
        slabinfo = {}
        # loop over all slab regions, return keep all slabs found
        for depth_file in self._depth_files:
            dip_file = pathlib.Path(str(depth_file).replace("dep", "dip"))
            strike_file = pathlib.Path(str(depth_file).replace("dep", "str"))
            error_file = pathlib.Path(str(depth_file).replace("dep", "unc"))
            if not error_file.is_file():
                error_file = None
            gslab = GridSlab(depth_file, dip_file, strike_file, error_file)
            tslabinfo = gslab.getSlabInfo(lat, lon)
            if not len(tslabinfo):
                continue
            else:
                depth = tslabinfo["depth"]
                if depth < deep_depth:
                    slabinfo = tslabinfo.copy()
                    deep_depth = depth
                elif np.isnan(depth) and "depth" not in slabinfo:
                    slabinfo = tslabinfo.copy()

        return slabinfo



def norm_angle(angle):
    """Normalize input angle to be between 0-360 degrees.

    Args:
        angle (float): Angle in degrees.
    Returns:
        float: Input angle normalized to between 0-360 degrees.
    """
    return angle % 360


class SubductionZone(object):
    def __init__(self, slab_params, tensor_params, depth, config):
        """Check where in the subduction zone this event is (crust, interface, slab).

        Args:
        slab_params (dict): Dictionary containing depth, strike and dip of
            the slab interface at a given location.
        tensor_params (dict): Dictionary containing the moment tensor
            parameters (six components, and nodal plane values.)
        depth (float): Event depth.
        config (dict): dict containing keys:
            DSTRIKE_INTERF - Threshold strike angle difference.
            DDIP_INTERF - Threshold dip angle difference.
            DLAMBDA - Threshold rake angle difference.
            DDEPTH_INTERF - Acceptable depth range around interface depth for interface
            event.
            DDEPTH_INTRA - intra-slab depth range.
        """
        self._dstrike = float(config["CONSTANTS"]["dstrike_interf"])
        self._ddip = float(config["CONSTANTS"]["ddip_interf"])
        self._dlambda = float(config["CONSTANTS"]["dlambda"])
        self._ddepth_interface = float(config["CONSTANTS"]["ddepth_interf"])
        self._ddepth_intraslab = float(config["CONSTANTS"]["ddepth_intra"])
        self._slab_params = slab_params.copy()
        if isinstance(tensor_params, dict):
            self._tensor_params = tensor_params.copy()
        else:
            self._tensor_params = None
        self._depth = depth

    def checkRupturePlane(self):
        """Compare moment tensor angles to slab angles, return True if similar.


        Returns:
            bool: Boolean value indicating if moment tensor is similar to slab.
        """
        if self._tensor_params is None:
            return False
        strike = self._slab_params["strike"] - 90
        a = self._tensor_params["P"]["azimuth"]
        b1 = norm_angle(strike) - self._dstrike
        b2 = norm_angle(strike) + self._dstrike
        b3 = norm_angle(strike) - self._dstrike
        b4 = norm_angle(strike) + self._dstrike

        if a > 270 and b1 < 90:
            b1 = b1 + 360
        if a > 270 and b2 < 90:
            b2 = b2 + 360
        if a > 270 and b3 < 90:
            b3 = b3 + 360
        if a > 270 and b4 < 90:
            b4 = b4 + 360

        if a < 90 and b1 > 270:
            b1 = b1 - 360
        if a < 90 and b2 > 270:
            b2 = b2 - 360
        if a < 90 and b3 > 270:
            b3 = b3 - 360
        if a < 90 and b4 > 270:
            b4 = b4 - 360

        c1 = a >= b1
        c2 = a <= b2
        c3 = a >= b3
        c4 = a <= b4

        m2a = (c1 and c2) or (c3 and c4)
        m2b = (
            self._tensor_params["P"]["plunge"] >= self._slab_params["dip"] - self._ddip
        ) and (
            self._tensor_params["P"]["plunge"] <= self._slab_params["dip"] + self._ddip
        )
        m2c1 = (self._tensor_params["NP1"]["rake"] > 90 - self._dlambda) and (
            self._tensor_params["NP1"]["rake"] < 90 + self._dlambda
        )
        m2c2 = (self._tensor_params["NP2"]["rake"] > 90 - self._dlambda) and (
            self._tensor_params["NP2"]["rake"] < 90 + self._dlambda
        )
        if m2a and m2b and (m2c1 or m2c2):
            return True
        else:
            return False

    def checkInterfaceDepth(self):
        """Check to see if the focal depth is within range of the interface depth.

        Returns:
            bool: True if event is close to slab interface depth, False otherwise.
        """
        c1 = self._depth >= self._slab_params["depth"] - self._ddepth_interface
        c2 = self._depth < self._slab_params["depth"] + self._ddepth_interface
        if c1 and c2:
            return True
        else:
            return False

    def checkSlabDepth(self, intraslabdepth):
        """Check to see if the depth is deeper than the interface.

        Args:
            intraslabdepth (float): Upper limit of intraslab events.

        Returns:
            True if depth is deeper than the slab, False if not.
        """
        c1 = self._depth >= self._slab_params["depth"] - self._ddepth_intraslab
        c2 = self._depth >= intraslabdepth
        if c1 and c2:
            return True
        else:
            return False



# stdlib imports
import logging
import pathlib

# third party libraries
from configobj import ConfigObj, flatten_errors, get_extra_values
from validate import ValidateError, Validator


def get_select_config(select_conf):
    select_spec = pathlib.Path(__file__).parent / "data" / "selectspec.conf"
    config = ConfigObj(str(select_conf), configspec=str(select_spec))
    validator = get_custom_validator()
    results = config.validate(validator, copy=True)
    if not isinstance(results, bool):
        return None, results
    return config, results


def get_custom_validator():
    """
    Returns a validator suitable for validating the ShakeMap config
    files.

    Returns:
        :class:`Validator`: A Validator object.

    """
    fdict = {
        "nanfloat_type": nanfloat_type,
        "nanfloat_list": nanfloat_list,
    }
    validator = Validator(fdict)
    return validator


def config_error(config, results):
    """
    Parse the results of a ConfigObj validation and log the errors.
    Throws a RuntimeError exception  upon completion if any errors or
    missing sections are encountered.

    Args:
        config (ConfigObj): The ConfigObj instance representing the
            parsed config.
        results (dict): The dictionary returned by the validation of
            the 'config' arguments.

    Returns:
        Nothing: Nothing

    Raises:
        RuntimeError: Should always raise this exception.
    """
    errs = 0
    for section_list, key, _ in flatten_errors(config, results):
        if key is not None:
            logging.error(
                'The "%s" key in the section "%s" failed validation'
                % (key, ", ".join(section_list))
            )
            errs += 1
        else:
            logging.error(
                f"The following section was missing:{', '.join(section_list)} "
            )
            errs += 1
    if errs:
        raise RuntimeError(
            "There %s %d %s in configuration."
            % ("was" if errs == 1 else "were", errs, "error" if errs == 1 else "errors")
        )


def check_extra_values(config, logger):
    """
    Checks the config and warns the user if there are any extra entries
    in their config file. This function is based on suggested usage in the
    ConfigObj manual.

    Args:
        config (ConfigObj): A ConfigObj instance.
        logger (logger): The logger to which to write complaints.

    Returns:
        Nothing: Nothing.
    """
    warnings = 0
    for sections, name in get_extra_values(config):
        # this code gets the extra values themselves
        the_section = config
        for section in sections:
            the_section = the_section[section]

        # the_value may be a section or a value
        the_value = the_section[name]

        section_or_value = "value"
        if isinstance(the_value, dict):
            # Sections are subclasses of dict
            section_or_value = "section"

        section_string = ", ".join(sections) or "top level"
        logger.warning(
            "Extra entry in section: %s: %s %r is not in spec."
            % (section_string, section_or_value, name)
        )
        warnings += 1
    if warnings:
        logger.warning(
            "The extra value(s) may be the result of deprecated "
            "entries or other changes to the config files; please "
            "check the conifg files in shakemap/data for the most "
            "up to date specs."
        )


def nanfloat_type(value):
    """
    Checks to see if value is a float, or NaN, nan, Inf, -Inf, etc.
    Raises a ValidateError exception on failure.

    Args:
        value (str): A string representing a float NaN or Inf.

    Returns:
        float: The input value converted to its float equivalent.

    """
    try:
        out = float(value)
    except ValueError:
        raise ValidateError(value)
    return out


def nanfloat_list(value, min):
    """
    Checks to see if value is a list of floats, including NaN and Inf.
    Raises a ValidateError exception on failure.

    Args:
        value (str): A string representing a list of floats.

    Returns:
        list: The input string converted to a list of floats.

    """
    min = int(min)
    if isinstance(value, str) and (value == "None" or value == "[]"):
        value = []
    if isinstance(value, str):
        value = [value]
    if isinstance(value, list) and not value:
        value = []
    if not isinstance(value, list):
        logging.error(f"'{value}' is not a list")
        raise ValidateError()
    if len(value) < min:
        logging.error("extent list must contain %i entries" % min)
        raise ValidateError()
    try:
        out = [float(a) for a in value]
    except ValueError:
        logging.error("%s is not a list of %i floats" % (value, min))
        raise ValidateError()
    return out



# stdlib imports
import json
import pathlib
import warnings
from datetime import datetime

# third party imports
import pyproj
from geopy.distance import distance as geodetic
from shapely.geometry import Point, shape
from shapely.ops import transform

# local imports
from strec.utils import get_config

CENTROID_MAX_DISTANCE = 6200  # approx length of largest subduction zone (SAM)
TECTONIC_REGIONS = {
    "stable": "DistanceToStable",
    "active": "DistanceToActive",
    "volcanic": "DistanceToVolcanic",
    "subduction": "DistanceToSubduction",
    "ocean": "DistanceToOceanic",
    "land": "DistanceToContinental",
    "backarc": "DistanceToBackarc",
}


def get_distance_to_shape(polygon_shape, clat, clon, dest_crs, regime):
    """Calculate distance to input geometry from lat/lon in given projection.

    Args:
        cutshape (shapely Geometry): Shape (usually Polygon) to calculate distance to.
        clat (float): Earthquake latitude.
        clon (float): Earthquake longitude.
        projection (Proj): Proj4 class defining an input projection.
    Returns:
        float: Distance in km from shape to lat/lon.
    """
    geo_crs = pyproj.CRS("EPSG:4326")
    pfunction = pyproj.Transformer.from_crs(geo_crs, dest_crs, always_xy=True).transform
    point = Point(clon, clat)
    proj_point = transform(pfunction, point)
    proj_shape = transform(pfunction, polygon_shape)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        distance = proj_shape.distance(proj_point) / 1000

    return distance


def get_distance_to_regime(lat, lon, regime):
    """Calculate distance from point to nearest polygon in a given regime.

    Args:
        lat (float): Earthquake latitude.
        lon (float): Earthquake longitude.
        regime (str): One of ["land", "ocean", "active", "stable", "volcanic",
                              "subduction", "backarc"]
    Returns:
        float: Distance (km) to nearest regime polygon.
    """
    config = get_config()
    longest_axis = float(config["DATA"]["longest_axis"])
    ortho_crs = pyproj.CRS(f"+proj=aeqd +lon_0={lon:.6f} +lat_0={lat:.6f} +ellps=WGS84")
    root = pathlib.Path(__file__).parent / "data"
    mindist = 1e9
    datafile = root / f"{regime}.geojson"
    with open(datafile, "rt") as f:
        jdict = json.load(f)
    for feature in jdict["features"]:
        geometry = shape(feature["geometry"])
        cx, cy = geometry.centroid.xy
        cdist = geodetic((lat, lon), (cy[0], cx[0])).km
        if cdist > longest_axis:
            continue
        dist = get_distance_to_shape(geometry, lat, lon, ortho_crs, regime)
        if dist < mindist:
            mindist = dist

    return mindist


def calc_distances(lat, lon):
    """Calculate distances from input lat/lon to nearest tectonic regime polygons.

    Args:
        lat (float): Earthquake latitude.
        lon (float): Earthquake longitude.
    Returns:
        dict: Dictionary of distances in km:
              - DistanceToActive
              - DistanceToStable
              - DistanceToSubduction
              - DistanceToVolcanic
              - DistanceToOceanic
              - DistanceToContinental
    """
    distances = {}
    regimes = ["active", "stable", "subduction", "volcanic", "ocean", "land", "backarc"]
    for regime in regimes:
        mindist = get_distance_to_regime(lat, lon, regime)
        distances[regime] = mindist

    for oldkey, newkey in TECTONIC_REGIONS.items():
        if oldkey not in distances:
            continue
        distances[newkey] = distances[oldkey]
        del distances[oldkey]

    return distances
